#include "../lib/mysql.hpp"


marcelb::mysql::MySQL::MySQL(const string _path, const string _username, const string _password, const string _db, const uint32_t _available, const periodical_engine _engine_type) {
   path = _path;
   username = _username;
   password = _password;
   database = _db;
   pool_size = _available > 0 ? _available : 1;
   engine_type = _engine_type;

   drv = get_mysql_driver_instance();
   connect_pool();

   if (engine_type == periodical_engine::internal) {
      periodic_engin = async(launch::async, [&](){
         while (run_engin) {
            usleep(MYSQL_PERIODIC_INTERNAL_TIME*1000);
            periodic_maintenance();
         }
         return;
      });  
   }
}


Connection* marcelb::mysql::MySQL::create_connection() {
   uint32_t trys = 0;
   bool status = true;
   Connection* new_con = NULL;

   while (connect_trys == unlimited ? status : (trys <= connect_trys && status)) {
      try {
         Connection* con_can = drv->connect(path, username, password);
         con_can->setSchema(database);
         status = !con_can->isValid();
         if (!status) {
            new_con = con_can;
         }
         else if (!con_can->isClosed()) {
            disconnect_connection(con_can);
         }
      }
      catch (const SQLException &error) {
         cout << error.what() << endl;
         // on_error -- ako se ikad impelementira pozovi ga ovdje!
         usleep(reconnectSleep);
         connect_trys == unlimited ? trys : trys++;
      }         
   }

   return new_con;
}

void marcelb::mysql::MySQL::connect_pool() {
   lock_guard<mutex> lock(io);
   for (uint32_t i=0; i<pool_size; i++) {
      Connection* connection = create_connection();
      connection_pool.push(connection);
   }
}

void marcelb::mysql::MySQL::disconnect_pool() {
   lock_guard<mutex> lock(io);
   for (uint32_t i=0; i<connection_pool.size(); i++) {
      Connection* connection = connection_pool.front();
      connection_pool.pop();
      disconnect_connection(connection) ;
   }
}

bool marcelb::mysql::MySQL::disconnect_connection(Connection* connection) {
   bool status = !connection->isClosed();

   if (status) {
      try {
         connection->close();
         status = !connection->isClosed();
      } 
      catch (const SQLException &error) {
         cout << error.what() << endl;
         status = true;
      }
   }

   else {
      status = false; // već je zatvorena
   }

   delete connection;
   return status;
}


/**
 * Broj pokušaja usljed povezivanja s bazom od 1 do unlimited;
*/

void marcelb::mysql::MySQL::set_connect_trys(const uint32_t _trys) {
   lock_guard<mutex> lock(io);
   connect_trys = _trys;
}

Connection* marcelb::mysql::MySQL::occupy_connection() {
   unique_lock<mutex> lock(io);
   while (connection_pool.empty()) {
      condition.wait(lock);
   }
   Connection *connection = connection_pool.front();
   connection_pool.pop();
   return connection;

   // while (true) {
   //    while(connection_pool.size()) {
   //       io.lock();
   //       Connection* connection = connection_pool.front();
   //       connection_pool.pop();
   //       if (connection->isValid()) {
   //          io.unlock();
   //          return connection;
   //       } 
   //       io.unlock();
   //    }
   //    usleep(1000);
   // }
}

void marcelb::mysql::MySQL::release_connection(Connection* connection) {
   lock_guard<std::mutex> lock(io);
   connection_pool.push(connection);
   condition.notify_one();
}

marcelb::mysql::MySQL::~MySQL() {
   if (engine_type == periodical_engine::internal) {
      run_engin = false;
      periodic_engin.get();
   } else {
      // ne bi bilo loše ubiti periodic nekako?!
      // iako disconnecta može periodic connect napraviti!!!
      run_engin = false;
   }
   
   disconnect_pool();
}


void marcelb::mysql::MySQL::periodic_maintenance() {
   for (size_t i = 0; i < pool_size && run_engin; i++) {
      Connection *conn = occupy_connection();
      if (conn->isValid()) {
         release_connection(conn);
      } else {
         if (!conn->isClosed()){
               conn->close();
         }
         Connection *n_conn = create_connection();
         release_connection(n_conn);
      }
   }
}